# -*- coding: utf-8 -*-
"""
Tests for the stats related utilities.
"""
import numpy as np

from nessai.utils.stats import effective_sample_size, rolling_mean


def test_ess():
    """Test the effective sample size"""
    log_w = np.zeros(10)
    np.testing.assert_almost_equal(effective_sample_size(log_w), 10)
    # Make the input array remains unchanged
    assert (log_w == 0.0).all()


def test_rolling_mean():
    """Test the rolling mean."""
    x = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
    expected = np.array([4.0 / 3.0, 2.0, 3.0, 4.0, 5.0, 17.0 / 3.0])
    out = rolling_mean(x, N=3)
    np.testing.assert_array_almost_equal(out, expected, decimal=15)



# -*- coding: utf-8 -*-
"""
Utilities related to statistics.
"""
import numpy as np
from scipy.special import logsumexp


def effective_sample_size(log_w):
    """Compute Kish's effective sample size.

    Parameters
    ----------
    log_w : array_like
        Log-weights.

    Returns
    -------
    float
        The effective sample size.
    """
    log_w = np.array(log_w)
    log_w -= logsumexp(log_w)
    n = np.exp(-logsumexp(2 * log_w))
    return n


def rolling_mean(x, N=10):
    """Compute the rolling mean with a given window size.

    Based on this answer from StackOverflow: \
        https://stackoverflow.com/a/47490020

    Parameters
    ----------
    x : :obj:`numpy.ndarray`
        Array of samples
    N : int
        Size of the window over which the rolling mean is computed.

    Returns
    -------
    :obj:`numpy.ndarray`
        Array containing the moving average.
    """
    # np.cumsum is faster but doesn't work with infs in the data.
    return np.convolve(
        np.pad(x, (N // 2, N - 1 - N // 2), mode="edge"),
        np.ones(N) / N,
        mode="valid",
    )



# -*- coding: utf-8 -*-
"""Different nested samplers available in nessai.

All samplers inherit from :py:obj:`nessai.samplers.base.BaseNestedSampler`.
"""
from .nestedsampler import NestedSampler

__all__ = [
    "NestedSampler",
]



# -*- coding: utf-8 -*-
"""
Main code that handles running and checkpoiting the sampler.
"""
import logging
import os
import signal
import sys
from typing import Optional

from .livepoint import live_points_to_dict
from .samplers.nestedsampler import NestedSampler
from .posterior import draw_posterior_samples
from .utils import configure_threads
from .utils.io import save_to_json, save_dict_to_hdf5


logger = logging.getLogger(__name__)


class FlowSampler:
    """
    Main class to handle running the nested sampler.

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User-defined model.
    output : str, optional
        Output directory
    resume : bool, optional
        If True try to resume the sampler is the resume file exists.
    resume_file : str, optional
        File to resume sampler from.
    weights_files : str, optional
        Weights files used to resume sampler that replaces the weights file
        saved internally.
    pytorch_threads : int
        Maximum number of threads to use for torch. If ``None`` torch uses all
        available threads.
    max_threads : int
        Deprecated and will be removed in a future release.
    signal_handling : bool
        Enable or disable signal handling.
    exit_code : int, optional
        Exit code to use when forceably exiting the sampler.
    close_pool : bool
        If True, the multiprocessing pool will be closed once the run method
        has been called. Disables the option in :code:`NestedSampler` if
        enabled.
    disable_vectorisation : bool
        Disable likelihood vectorisation. Overrides the value of
        :py:attr:`nessai.model.Model.allow_vectorised`.
    likelihood_chunksize : Optional[int]
        Chunksize used when evaluating a vectorised likelihood. Overrides the
        of :py:attr:`nessai.model.Model.likelihood_chunksize`. Set to None to
        evaluate the likelihood with all available points.
    allow_multi_valued_likelihood : Optional[bool]
        Allow for a multi-valued likelihood function that will return different
        likelihood values for the same point in parameter space. See
        :py:attr:`nessai.model.Model.allow_multi_valued_likelihood` for more
        details.
    result_extension : str
        Extension used when saving the result format. Defaults to HDF5, but
        also supports JSON.
    kwargs :
        Keyword arguments passed to
        :obj:`~nessai.samplers.nestedsampler.NestedSampler`.
    """

    def __init__(
        self,
        model,
        output=os.getcwd(),
        resume=True,
        resume_file="nested_sampler_resume.pkl",
        weights_file=None,
        signal_handling=True,
        exit_code=130,
        pytorch_threads=1,
        max_threads=None,
        close_pool=True,
        disable_vectorisation=False,
        likelihood_chunksize=None,
        allow_multi_valued_likelihood=None,
        result_extension="hdf5",
        **kwargs,
    ):

        configure_threads(
            max_threads=max_threads,
            pytorch_threads=pytorch_threads,
        )

        self.exit_code = exit_code
        self.close_pool = close_pool

        self.result_extension = result_extension

        if disable_vectorisation:
            logger.warning(
                "Overriding value of `allow_vectorised` in the model"
            )
            model.allow_vectorised = False

        if likelihood_chunksize:
            model.likelihood_chunksize = likelihood_chunksize

        if allow_multi_valued_likelihood is not None:
            model.allow_multi_valued_likelihood = allow_multi_valued_likelihood

        self.output = os.path.join(output, "")
        os.makedirs(self.output, exist_ok=True)
        self.save_kwargs(kwargs)

        model.configure_pool(
            n_pool=kwargs.pop("n_pool", None), pool=kwargs.pop("pool", None)
        )

        if resume:
            if not resume_file:
                raise RuntimeError(
                    "`resume_file` must be specified if resume=True. "
                    f"Current value: {resume_file}"
                )
            if not any(
                (
                    os.path.exists(os.path.join(self.output, f))
                    for f in [resume_file, resume_file + ".old"]
                )
            ):
                logger.warning("No files to resume from, starting sampling")
                self.ns = NestedSampler(
                    model,
                    output=self.output,
                    resume_file=resume_file,
                    close_pool=not self.close_pool,
                    **kwargs,
                )
            else:
                try:
                    self.ns = NestedSampler.resume(
                        os.path.join(self.output, resume_file),
                        model,
                        flow_config=kwargs.get("flow_config"),
                        weights_file=weights_file,
                    )
                except (FileNotFoundError, RuntimeError) as e:
                    logger.error(
                        f"Could not load resume file from: {self.output} "
                        f"with error {e}"
                    )
                    try:
                        resume_file += ".old"
                        self.ns = NestedSampler.resume(
                            os.path.join(self.output, resume_file),
                            model,
                            flow_config=kwargs.get("flow_config"),
                            weights_file=weights_file,
                        )
                    except RuntimeError as e:
                        logger.error(
                            "Could not load old resume "
                            f"file from: {self.output}"
                        )
                        raise RuntimeError(
                            "Could not resume sampler " f"with error: {e}"
                        )
        else:
            self.ns = NestedSampler(
                model,
                output=self.output,
                resume_file=resume_file,
                close_pool=not self.close_pool,
                **kwargs,
            )

        if signal_handling:
            try:
                signal.signal(signal.SIGTERM, self.safe_exit)
                signal.signal(signal.SIGINT, self.safe_exit)
                signal.signal(signal.SIGALRM, self.safe_exit)
            except AttributeError:
                logger.error("Cannot set signal attributes on this system")
        else:
            logger.warning(
                "Signal handling is disabled. nessai will not automatically "
                "checkpoint when exitted."
            )

    @property
    def log_evidence(self):
        """Return the most recent log evidence"""
        return self.logZ

    @property
    def log_evidence_error(self):
        """Return the most recent log evidence error"""
        return self.logZ_error

    def run(
        self,
        plot=True,
        plot_indices=True,
        plot_posterior=True,
        plot_logXlogL=True,
        save=True,
        posterior_sampling_method=None,
        close_pool=None,
    ):
        """Run the nested sampler.

        Parameters
        ----------
        plot : bool
            Toggle all plots produced once the sampler has converged.
        plot_indices : bool
            Toggle the insertion indices plot.
        plot_posterior : bool
            Toggle the posterior distribution plot.
        plot_logXlogL : bool
            Toggle the log-prior volume vs log-likelihood plot.
        save : bool, optional
            Toggle automatic saving of results
        posterior_sampling_method : str, optional
            Method used for drawing posterior samples. Defaults to rejection
            sampling.
        close_pool : bool, optional
            Boolean to indicated if the pool should be closed at the end of the
            run function. If False, the user must manually close the pool. If
            specified, this value overrides the value passed when initialising
            the FlowSampler class.
        """
        if close_pool is None:
            close_pool = self.close_pool
        if posterior_sampling_method is None:
            posterior_sampling_method = "rejection_sampling"

        self.ns.initialise()
        self.logZ, self.nested_samples = self.ns.nested_sampling_loop()
        self.logZ_error = self.ns.state.log_evidence_error
        logger.info((f"Total sampling time: {self.ns.sampling_time}"))

        logger.info("Starting post processing")
        logger.info("Computing posterior samples")
        self.posterior_samples = draw_posterior_samples(
            self.nested_samples,
            log_w=self.ns.state.log_posterior_weights,
            method=posterior_sampling_method,
        )
        logger.info(
            f"Returned {self.posterior_samples.size} " "posterior samples"
        )

        if save:
            self.save_results(
                os.path.join(self.output, "result"),
                extension=self.result_extension,
            )

        if plot:
            from nessai import plot

            if plot_posterior:
                plot.plot_live_points(
                    self.posterior_samples,
                    filename=os.path.join(
                        self.output, "posterior_distribution.png"
                    ),
                )
            if plot_indices:
                plot.plot_indices(
                    self.ns.insertion_indices,
                    self.ns.nlive,
                    filename=os.path.join(
                        self.output, "insertion_indices.png"
                    ),
                )

            if plot_logXlogL:
                self.ns.state.plot(os.path.join(self.output, "logXlogL.png"))

        if close_pool:
            self.ns.close_pool()

    def save_kwargs(self, kwargs: dict) -> None:
        """
        Save the dictionary of keyword arguments used.

        Uses an encoder class to handle numpy arrays.

        Parameters
        ----------
        kwargs : dict
            Dictionary of kwargs to save.
        """
        save_to_json(kwargs.copy(), os.path.join(self.output, "config.json"))

    def save_results(
        self, filename: str, extension: Optional[str] = None
    ) -> None:
        """
        Save the results from sampling to a specific results file.

        Parameters
        ----------
        filename : str
            Name of file to save results to.
        extension : Optional[str]
            File extension to used. If not specified, it will be inferred from
            the filename.
        """
        d = self.ns.get_result_dictionary()
        d["posterior_samples"] = self.posterior_samples

        ext = os.path.splitext(filename)[1].lstrip(".")
        if extension is None:
            if ext == "":
                raise RuntimeError(
                    "Must specify file extension if not present in filename!"
                )
            else:
                extension = ext
        elif ext == "":
            filename = ".".join([filename, extension])

        if extension == "json":
            d["posterior_samples"] = live_points_to_dict(
                d["posterior_samples"]
            )
            save_to_json(d, filename)
        elif extension in ["hdf5", "h5"]:
            save_dict_to_hdf5(d, filename)
        else:
            raise RuntimeError(f"Unknown file extension: {extension}")

    def terminate_run(self, code=None):
        """Terminate a sampling run.

        Parameters
        ----------
        code : int, optional
            Code passed to :code:`close_pool`
        """
        logger.warning("Terminating run")
        self.ns.close_pool(code=code)
        self.ns.checkpoint()

    def safe_exit(self, signum=None, frame=None):
        """
        Safely exit. This includes closing the multiprocessing pool.
        """
        logger.warning(f"Trying to safely exit with code {signum}")
        self.terminate_run(code=signum)
        logger.warning(f"Exiting with code: {self.exit_code}")
        sys.exit(self.exit_code)



# -*- coding: utf-8 -*-
"""nessai: Nested sampling with Artificial Intelligence

nessai is a nested sampling algorithm for Bayesian Inference that incorporates
normalising flows. It is designed for applications where the Bayesian
likelihood is computationally expensive.
"""
import logging

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:  # for Python < 3.8
    from importlib_metadata import version, PackageNotFoundError


try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    __version__ = "unknown"

logging.getLogger(__name__).addHandler(logging.NullHandler())



# -*- coding: utf-8 -*-
"""
Tests for bilbyutils
"""
from unittest.mock import patch
from nessai.flowsampler import FlowSampler
from nessai.utils.bilbyutils import (
    get_all_kwargs,
    get_run_kwargs_list,
    _get_standard_methods,
)


def test_get_standard_methods():
    """Assert a list of methods is returned"""
    out = _get_standard_methods()
    assert len(out) == 5


def test_get_all_kwargs():
    """Assert the correct dictionary is returned.

    Positional arguments should be ignored.
    """

    def func0(a, b, c=2, d=None):
        pass

    def func1(e, f, g=3, h=True):
        pass

    expected = dict(c=2, d=None, g=3, h=True)

    with patch(
        "nessai.utils.bilbyutils._get_standard_methods",
        return_value=[func0, func1],
    ):
        out = get_all_kwargs()

    assert out == expected


def test_get_run_kwargs_list():
    """Assert the correct list is returned"""

    def func(a, b=1, c=None):
        pass

    expected = ["b", "c"]

    with patch.object(FlowSampler, "run", func):
        out = get_run_kwargs_list()

    assert out == expected



# -*- coding: utf-8 -*-
"""
Tests for rescaling functions
"""
import multiprocessing
from multiprocessing.dummy import Pool
import pytest
from unittest.mock import MagicMock, patch

from nessai.utils.multiprocessing import (
    check_multiprocessing_start_method,
    initialise_pool_variables,
    get_n_pool,
    log_likelihood_wrapper,
)


def test_pool_variables():
    """Assert initialising the variables and calling the likelihood work"""
    model = MagicMock()
    model.log_likelihood = lambda x: x
    initialise_pool_variables(model)
    pool = Pool(1)
    out = pool.map(log_likelihood_wrapper, [1, 2, 3])
    pool.close()
    pool.terminate()
    assert out == [1, 2, 3]

    # Reset to the default value
    initialise_pool_variables(None)


def test_check_multiprocessing_start_method():
    """Test check multiprocessing start method passes for 'fork'"""
    with patch("multiprocessing.get_start_method", return_value="fork"):
        check_multiprocessing_start_method()


@pytest.mark.parametrize("method", ["spawn", "forkserver"])
def test_check_multiprocessing_start_method_error(method):
    """Assert an error is raised if the start method is not fork."""
    error_msg = r"nessai only supports multiprocessing using the 'fork' .*"
    with patch(
        "multiprocessing.get_start_method", return_value=method
    ), pytest.raises(RuntimeError, match=error_msg):
        check_multiprocessing_start_method()


@pytest.mark.integration_test
@pytest.mark.skip_on_windows
def test_check_multiprocessing_start_method_integration():
    """Integration test for checking the start method."""
    mp = multiprocessing.get_context("fork")
    with patch("multiprocessing.get_start_method", mp.get_start_method):
        check_multiprocessing_start_method()


@pytest.mark.integration_test
def test_check_multiprocessing_start_method_error_integration():
    """Integration test for checking the start method raises an error."""
    mp = multiprocessing.get_context("spawn")
    error_msg = r"nessai only supports multiprocessing using the 'fork' .*"
    with patch(
        "multiprocessing.get_start_method", mp.get_start_method
    ), pytest.raises(RuntimeError, match=error_msg):
        check_multiprocessing_start_method()


def test_model_error():
    """Assert an error is raised in the global variables have not been \
        initialised
    """
    model = MagicMock()
    model.log_likelihood = lambda x: x
    pool = Pool(1)
    with pytest.raises(AttributeError) as excinfo:
        pool.map(log_likelihood_wrapper, [1, 2, 3])
    assert "'NoneType' object has no attribute 'log_likelihood'" in str(
        excinfo.value
    )
    pool.close()
    pool.terminate()


def test_get_n_pool_processes():
    """ "Assert the correct value is returned if the pool has a `_processes`
    attribute.
    """
    pool = MagicMock(spec=Pool)
    pool._processes = 4
    assert get_n_pool(pool) == 4


def test_get_n_pool_ray():
    """ "Assert the correct value is returned if the pool has a `_actor_pool`
    attribute (e.g. ray.util.multiprocessing.Pool).
    """
    pool = MagicMock()
    del pool._processes
    pool._actor_pool = 4 * [0]
    assert get_n_pool(pool) == 4


def test_get_n_pool_unknown():
    """ "Assert None is returned if the type of pool is not known."""
    pool = MagicMock()
    del pool._processes
    del pool._actor_pool
    assert get_n_pool(pool) is None



# -*- coding: utf-8 -*-
"""
Utilities for configuring FlowModel.
"""
from .config import DEFAULT_MODEL_CONFIG, DEFAULT_FLOW_CONFIG
from ..flows.utils import get_n_neurons


def update_config(d):
    """
    Update the configuration dictionary to include the defaults.

    Notes
    -----
    The default configuration is specified in :py:mod:`nessai.flowmodel.config`


    The kwargs can contain any additional keyword arguments that are specific
    to the type of flow being used.

    Parameters
    ----------
    d : dict
        Dictionary with configuration

    Returns
    -------
    dict
        Dictionary with updated default configuration
    """
    default = DEFAULT_FLOW_CONFIG.copy()
    default_model = DEFAULT_MODEL_CONFIG.copy()

    if d is None:
        default["model_config"] = default_model
    else:
        if not isinstance(d, dict):
            raise TypeError(
                "Must pass a dictionary to update the default "
                "trainer settings"
            )
        else:
            default.update(d)
            default_model.update(d.get("model_config", {}))
            default_model["n_neurons"] = get_n_neurons(
                n_neurons=default_model.get("n_neurons"),
                n_inputs=default_model.get("n_inputs"),
            )
            default["model_config"] = default_model

    if default["noise_type"] is not None and default["noise_scale"] is None:
        raise RuntimeError(
            "`noise_scale` must be specified when `noise_type` is given."
        )
    if isinstance(default["noise_scale"], float):
        if default["noise_type"] is None:
            default["noise_type"] = "constant"
    elif default["noise_scale"] is not None:
        raise TypeError(
            "`noise_scale` must be a float. "
            f"'Got type: {type(default['noise_scale'])}"
        )

    return default



# -*- coding: utf-8 -*-
"""
Utilities related to multiprocessing.
"""
import logging
import multiprocessing

_model = None
logger = logging.getLogger(__name__)


def get_n_pool(pool):
    """Determine the number of processes in a multiprocessing pool.

    Parameters
    ----------
    pool : object
        Multiprocessing pool or similar.

    Returns
    -------
    int or None
        Number of processes. Returns None if number could not be determined.
    """
    try:
        n_pool = pool._processes
    except AttributeError:
        try:
            n_pool = len(pool._actor_pool)
        except AttributeError:
            n_pool = None
            logger.warning(
                "Could not determine number of processes in pool of type: "
                f"{type(pool)}."
            )
    return n_pool


def check_multiprocessing_start_method():
    """Check the multiprocessing start method.

    Raise an error if the start method is not `fork`.
    """
    start_method = multiprocessing.get_start_method()
    if start_method != "fork":
        raise RuntimeError(
            "nessai only supports multiprocessing using the 'fork' start "
            f"method. Actual start method is: {start_method}. See the "
            "multiprocessing documentation for more details."
        )


def initialise_pool_variables(model):
    """Prepare the model for use with a multiprocessing pool.

    Makes a global copy of the model. Should be called before initialising
    a pool or passed to the :code:`initializer` argument with the model as one
    of the :code:`initargs`.

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        Model to be copied to a global variable.
    """
    global _model
    _model = model


def log_likelihood_wrapper(x):
    """Wrapper for the log-likelihood for use with multiprocessing.

    Should be used alongside
    :py:func:`nessai.utils.multiprocessing.initialise_pool_variables`

    Parameters
    ----------
    x : :obj:`numpy.ndarray`
        Array of samples.

    Returns
    -------
    :obj:`numpy.ndarray`
        Array of log-likelihoods.
    """
    return _model.log_likelihood(x)



# -*- coding: utf-8 -*-
"""
Functions related to computing the evidence.
"""
from abc import ABC, abstractmethod
import logging

import numpy as np
import matplotlib.pyplot as plt
from scipy.special import logsumexp

from .plot import nessai_style

logger = logging.getLogger(__name__)


def logsubexp(x, y):
    """
    Helper function to compute the exponential
    of a difference between two numbers

    Computes: ``x + np.log1p(-np.exp(y-x))``

    Parameters
    ----------
    x, y : float or array_like
        Inputs
    """
    if np.any(x < y):
        raise RuntimeError(
            "cannot take log of negative number " f"{str(x)!s} - {str(y)!s}"
        )

    return x + np.log1p(-np.exp(y - x))


def log_integrate_log_trap(log_func, log_support):
    """
    Trapezoidal integration of given log(func). Returns log of the integral.

    Parameters
    ----------
    log_func : array_like
        Log values of the function to integrate over.
    log_support : array_like
        Log prior-volumes for each value.

    Returns
    -------
    float
        Log of the result of the integral.
    """
    log_func_sum = np.logaddexp(log_func[:-1], log_func[1:]) - np.log(2)
    log_dxs = logsubexp(log_support[:-1], log_support[1:])

    return logsumexp(log_func_sum + log_dxs)


class _BaseNSIntegralState(ABC):
    """Base class for the nested sampling integral."""

    @property
    @abstractmethod
    def log_evidence(self):
        """The current log-evidence."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def log_evidence_error(self):
        """The current error on the log-evidence."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def log_posterior_weights(self):
        """The log-posterior weights."""
        raise NotImplementedError()

    @property
    def effective_n_posterior_samples(self):
        """Kish's effective sample size for the posterior weights.

        Returns
        -------
        float
            The effective sample size. Returns zero if the posterior weights
            are empty.
        """
        log_p = self.log_posterior_weights
        if not len(log_p):
            return 0
        log_p -= logsumexp(log_p)
        n = np.exp(-logsumexp(2 * log_p))
        return n


class _NSIntegralState(_BaseNSIntegralState):
    """
    Stores the state of the nested sampling integrator

    Parameters
    ----------
    nlive : int
        Number of live points
    track_gradients : bool, optional
        If true the gradient of the change in logL w.r.t logX is saved each
        time `increment` is called.
    expectation : str, {logt, t}
        Method used to compute the expectation value for the shrinkage t.
        Choose between log <t> or <log t>. Defaults to <log t>.
    """

    def __init__(self, nlive, track_gradients=True, expectation="logt"):
        self.base_nlive = nlive
        self.track_gradients = track_gradients

        if expectation.lower() not in ["t", "logt"]:
            raise ValueError(
                f"Expectation must be t or logt, got: {expectation}"
            )
        self.expectation = expectation.lower()

        # Initial state of the integral
        self.logZ = -np.inf
        self.oldZ = -np.inf
        self.logw = 0
        self.info = [0.0]
        # Initially contain all the prior volume
        self.logLs = [-np.inf]  # Likelihoods sampled
        self.log_vols = [0.0]  # Volumes enclosed by contours
        self.nlive = []
        self.gradients = [0]

    @property
    def log_evidence(self):
        """The current log-evidence."""
        return self.logZ

    @property
    def log_evidence_error(self):
        """The current error on the log-evidence."""
        return np.sqrt(self.info[-1] / self.base_nlive)

    def increment(self, logL, nlive=None):
        """
        Increment the state of the evidence integrator
        Simply uses rectangle rule for initial estimate
        """
        if logL <= self.logLs[-1]:
            logger.warning(
                "NS integrator received non-monotonic logL."
                f"{self.logLs[-1]:.5f} -> {logL:.5f}"
            )
        if nlive is None:
            nlive = self.base_nlive

        self.nlive.append(nlive)
        oldZ = self.logZ
        if self.expectation == "logt":
            # <logt> approx -1 / N
            logt = -1.0 / nlive
        else:
            # <t> = N / (N + 1)
            logt = -np.log1p(1 / nlive)
        Wt = self.logw + logL + np.log1p(-np.exp(logt))
        self.logZ = np.logaddexp(self.logZ, Wt)
        # Update information estimate
        if np.isfinite(oldZ) and np.isfinite(self.logZ) and np.isfinite(logL):
            info = (
                np.exp(Wt - self.logZ) * logL
                + np.exp(oldZ - self.logZ) * (self.info[-1] + oldZ)
                - self.logZ
            )
            self.info.append(info)

        # Update history
        self.logw += logt
        self.logLs.append(logL)
        self.log_vols.append(self.logw)
        if self.track_gradients:
            self.gradients.append(
                (self.logLs[-1] - self.logLs[-2])
                / (self.log_vols[-1] - self.log_vols[-2])
            )

    def finalise(self):
        """
        Compute the final evidence with more accurate integrator
        Call at end of sampling run to refine estimate
        """
        # Trapezoidal rule
        # Extra point represents X=0 and assume max(L) = L[-1]
        self.logZ = log_integrate_log_trap(
            np.array(self.logLs + [self.logLs[-1]]),
            np.array(self.log_vols + [np.NINF]),
        )
        return self.logZ

    @nessai_style()
    def plot(self, filename=None):
        """
        Plot the logX vs logL

        Parameters
        ----------
        filename : str, optional
            Filename name for saving the figure. If not specified the figure
            is returned.
        """
        fig = plt.figure()
        plt.plot(self.log_vols, self.logLs)
        plt.title(
            f"log Z={self.logZ:.2f} "
            f"H={self.info[-1] * np.log2(np.e):.2f} bits"
        )
        plt.grid(which="both")
        plt.xlabel("log prior-volume")
        plt.ylabel("log-likelihood")
        plt.xlim([self.log_vols[-1], self.log_vols[0]])

        if filename is not None:
            fig.savefig(filename, bbox_inches="tight")
            plt.close()
            logger.debug(f"Saved nested sampling plot as {filename}")
        else:
            return fig

    @property
    def log_posterior_weights(self):
        """Compute the log-posterior weights."""
        log_L = np.array(self.logLs + [self.logLs[-1]])
        log_vols = np.array(self.log_vols + [np.NINF])
        log_Z = log_integrate_log_trap(log_L, log_vols)
        log_w = logsubexp(log_vols[:-1], log_vols[1:])
        log_post_w = log_L[1:-1] + log_w[:-1] - log_Z
        return log_post_w



# -*- coding: utf-8 -*-
"""
Utilities to make interfacing with bilby easier.
"""
from inspect import signature
from typing import List, Callable


def _get_standard_methods() -> List[Callable]:
    """Get a list of the methods used by the standard sampler"""
    from ..flowsampler import FlowSampler
    from ..proposal import AugmentedFlowProposal, FlowProposal
    from ..samplers import NestedSampler

    methods = [
        FlowSampler.run,
        AugmentedFlowProposal,
        FlowProposal,
        NestedSampler,
        FlowSampler,
    ]
    return methods


def get_all_kwargs() -> dict:
    """Get a dictionary of all possible kwargs and their default values.

    Returns
    -------
    Dictionary of kwargs and their default values.
    """
    methods = _get_standard_methods()
    kwargs = {}
    for m in methods:
        kwargs.update(
            {
                k: v.default
                for k, v in signature(m).parameters.items()
                if v.default is not v.empty
            }
        )
    return kwargs


def get_run_kwargs_list() -> List[str]:
    """Get a list of kwargs used in the run method"""
    from ..flowsampler import FlowSampler

    method = FlowSampler.run

    run_kwargs_list = [
        k
        for k, v in signature(method).parameters.items()
        if v.default is not v.empty
    ]
    return run_kwargs_list



# -*- coding: utf-8 -*-
"""
Utilities related to managing threads used by nessai.
"""
import logging
import warnings

import torch

logger = logging.getLogger(__name__)


def configure_threads(pytorch_threads=None, max_threads=None):
    """Configure the number of threads available.

    This is necessary when using PyTorch on the CPU as by default it will use
    all available threads.

    Notes
    -----
    Uses ``torch.set_num_threads``. If pytorch threads is None but other
    arguments are specified then the value is inferred from them.

    Parameters
    ----------
    pytorch_threads: int, optional
        Maximum number of threads for PyTorch on CPU. If None, pytorch will
        use all available threads.
    max_threads: int, optional
        Ignored. Deprecated starting nessai 0.7.0.
    """
    if max_threads:
        msg = (
            "`max_threads` is deprecated and will be removed in a future "
            "release. Use `pytorch_threads` to set the number of threads for "
            "pytorch and `n_pool` to set the number of cores for "
            "paralellisation."
        )
        warnings.warn(msg, FutureWarning)
    if pytorch_threads:
        logger.debug(
            f"Setting maximum number of PyTorch threads to {pytorch_threads}"
        )
        torch.set_num_threads(pytorch_threads)



# -*- coding: utf-8 -*-
"""Tests related to drawing new points from the pool."""
import numpy as np
import pytest
from unittest.mock import MagicMock, Mock

from nessai.livepoint import dict_to_live_points
from nessai.proposal import FlowProposal


def test_draw_populated(proposal):
    """Test the draw method if the proposal is already populated"""
    proposal.populated = True
    proposal.samples = np.arange(3)
    proposal.indices = list(range(3))
    out = FlowProposal.draw(proposal, None)
    assert out == proposal.samples[2]
    assert proposal.indices == [0, 1]


def test_draw_populated_last_sample(proposal):
    """Test the draw method if the proposal is already populated but there
    is only one sample left.
    """
    proposal.populated = True
    proposal.samples = np.arange(3)
    proposal.indices = [0]
    out = FlowProposal.draw(proposal, None)
    assert out == proposal.samples[0]
    assert proposal.indices == []
    assert proposal.populated is False


@pytest.mark.parametrize("update", [False, True])
def test_draw_not_populated(proposal, update, wait):
    """Test the draw method when the proposal is not populated"""
    import datetime

    proposal.populated = False
    proposal.poolsize = 100
    proposal.population_time = datetime.timedelta()
    proposal.samples = None
    proposal.indices = []
    proposal.update_poolsize = update
    proposal.update_poolsize_scale = MagicMock()
    proposal.ns_acceptance = 0.5

    def mock_populate(*args, **kwargs):
        wait()
        proposal.populated = True
        proposal.samples = np.arange(3)
        proposal.indices = list(range(3))

    proposal.populate = MagicMock(side_effect=mock_populate)

    out = FlowProposal.draw(proposal, 1.0)

    assert out == 2
    assert proposal.populated is True
    assert proposal.population_time.total_seconds() > 0.0

    proposal.populate.assert_called_once_with(1.0, N=100)

    assert proposal.update_poolsize_scale.called == update


def test_test_draw(proposal):
    """
    Test the method that tests the draw and populate methods when running.
    """
    test_point = dict_to_live_points(
        {"x": 1, "y": 2, "logP": -0.5}, non_sampling_parameters=False
    )
    new_point = dict_to_live_points(
        {"x": 3, "y": 4, "logP": -0.5}, non_sampling_parameters=False
    )
    proposal.model = Mock()
    proposal.model.new_point = MagicMock(return_value=test_point)
    proposal.model.log_prior = MagicMock(return_value=-0.5)
    proposal.populate = MagicMock()
    proposal.draw = MagicMock(return_value=new_point)
    proposal.reset = MagicMock()

    FlowProposal.test_draw(proposal)

    proposal.populate.assert_called_once_with(
        test_point, N=1, plot=False, r=1.0
    )
    proposal.reset.assert_called_once()
    proposal.draw.assert_called_once_with(test_point)


@pytest.mark.timeout(10)
@pytest.mark.flaky(reruns=3)
@pytest.mark.integration_test
def test_test_draw_integration(model, tmpdir):
    """Integration test for the test draw method"""
    proposal = FlowProposal(
        model,
        poolsize=10,
        output=tmpdir.mkdir("test_draw"),
        volume_fraction=0.5,
    )
    proposal.initialise()
    proposal.test_draw()
